//
// Gadget Styles
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <guit.hpp>
#include <gcomputedstyle.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gstyle.hpp>
#include <gadgetstyle.hpp>
using namespace std;
namespace guit {

GadgetStyle::GadgetStyle(GString const& ids)
: realized_(false), names_(new GString(ids)) {}

GadgetStyle::~GadgetStyle() {
  if (realized_) delete [] ids_;
  else delete names_;
}

GString GadgetStyle::typeName() const {return "GadgetStyle";}

GProp* GadgetStyle::findProp( Gadget const& g, GPropType& cl) {
  if (!realized_) realize(g);
  if (!ids_) return nullptr;

  GProp* prop{};
  for (auto* it = ids_; *it != 0; ++it) {  // 0 ended
    // pas de break car on cherche la derniere valeur !
    if (GStyle* s = GStyleBase::instance().findStyle(*it)) {
      if (auto* p = s->findProp(cl)) prop = p;
    }
  }
  return prop;
}

// can't be done in constr beacuse class is always Gadget
void GadgetStyle::realize(Gadget const& g) {
  if (realized_ || !names_) return;
  realized_ = true;

  GStrings tab;
  auto n = strsplit(*names_, tab);

  delete names_;  // not useful any more
  ids_ = nullptr;
  if (n == 0) return;

  auto tname = g.typeName();
  //GId tmp[2 * n];  // invalid on windows
  std::vector<GId> tmp(2 * n);
  size_t size = 0;

  for (auto& name : tab) {
    if (name[0] == ':') name.erase(0,1);
    if (auto id = GCore::findId(name)) tmp[size++] = id;
    if (auto id = GCore::findId(tname + name)) tmp[size++] = id;
  }
  if (size == 0) return;

  ids_ = new GId[size + 1];  // 0 ended
  for (size_t k = 0; k < size; ++k) ids_[k] = tmp[k];
  ids_[size] = 0;  // 0 terminated
}

// - - - - - -

static char findTokenBegin(GString const& text, size_t& k) {
  while (k < text.size() && ::isspace(text[k])) k++;

  if (k >= text.size()) return 0;
  else if (text[k]=='.' || text[k]=='#' || text[k]=='/' || text[k]==':' || text[k]=='{')
    return text[k];
  else return 0;
}

static void findTokenEnd(GString const& text, size_t& k) {
  size_t k0 = k;
  while (k < text.size() && !::isspace(text[k])) k++;
  if (k > k0) k--;
}

static void findBracketEnd(GString const& text, size_t& k) {
  while (k < text.size() && text[k]!='}') k++;
}
  
bool GadgetStyle::initStyle(Gadget& g, GString const& spec, GString& title) {
  bool stat = true;
  GString instyle;
  bool css_found = false, other_found = false;
  size_t sel_start = 0, sel_end = 0;
  size_t k = 0;
  const auto len = spec.size();

  while (k < len) {
    size_t begin = k;

    if (k+1 >= len) {
      title = spec.substr(begin);
      break;
    }

    char c = findTokenBegin(spec, k);
    if (c == 0) {
      if (css_found || other_found) title = spec.substr(begin+1);
      else title = spec.substr(begin);
      break;
    }

    switch (c) {
      case '.':
      case '#':
      case '/':
      case ':':
        if (other_found) {
          gerror("Gadget Argument","CSS classes must be first: "+spec);
          stat = false;
        }
        if (!css_found) sel_start = k;
        css_found = true;
        findTokenEnd(spec, k);
        sel_end = k;
        k++;
        break;

      case '{': {
        other_found = true;
        size_t start = k;
        findBracketEnd(spec, k);
        if (k >= start+3) instyle = spec.substr(start+1, k-start-1);
        k++;
        if (k < spec.size()) title = spec.substr(k);
        goto END;
      } break;
        /*
         case '&': {
         other_found = true;
         ssize_t start = k;
         findTokenEnd(spec, k);
         if (k > start) this->hotkey = spec.substr(start+1, k-start);
         k++;
         } break;
         default:
         other_found = true;
         break;
         */
    }
  }

END:
  // remove first space
  if (!title.empty() && ::isspace(title[0])) title.erase(0,1);

  // the gadget has a CSS, inits the GadgetStyle and creates the props.
  g.omodes_.initialized = 0;
  if (css_found && sel_end > sel_start) {
    g.style_ = new GadgetStyle(spec.substr(sel_start, sel_end-sel_start+1));
  }
  else g.style_ = nullptr;

  // there is an inline style {...} => add corresponding props
  if (!instyle.empty()) {
    if (!parseProps(&g, nullptr, instyle, ";")) stat = false;
  }
  return stat;
}


bool GadgetStyle::parseProps(Gadget* g, GPropList* pl, GString const& style,
                             GString const& delim) {
  GStrings tab;
  strsplit(style, tab, delim, 0, SplitMode::RemQuotes);

  for (auto& it : tab) {
    if (it.empty()) continue;
    GString propname, propval;
    auto count = strsplit(it, propname, propval, ":", SplitMode::RemQuotes);

    if (count == 1) {
      gerror("GadgetStyle","no value after : in {" + style + "}");
      return false;
    }
    else if (count != 2) {
      gerror("GadgetStyle",": is missing in {" + style + "}");
      return false;
    }
    else if (auto* p = GPropType::createInstance(propname, propval, true)) {
      if (g) g->add(*p);
      // attention la derniere definition ecrase les precedentes !
      else pl->addReplace(*p, nullptr);
    }
    else {
      gerror("GadgetStyle","Unknown Prop '" + propname + "' in {" + style + "}");
      return false;
    }
  }
  return true;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static void computeStyle2(GRender* parrender,
                          std::vector<Gadget*> const& parents, size_t index,
                          Gadget const* searched, GRender& result)
{
  if (index < 0) {
    gerror("GComputedStyle","could not compute style");
    return;
  }
  GRender r(*parents[index], parrender);
  if (parents[index] == searched) {
    result = r;
    return;
  }
  else computeStyle2(&r, parents, index-1, searched, result);
}


GComputedStyle::~GComputedStyle() {
  delete render_;
}

GComputedStyle::GComputedStyle(Gadget& obj)
: render_(new GRender) {
  std::vector<Gadget*> parents;
  Gadget* g = &obj;
  Gadget* toplevel{};

  while (g) {
    parents.push_back(g);
    if (g->isTopLevel()) {toplevel = g; break;}
    g = g->parent_;
  }
  if (toplevel) computeStyle2(nullptr, parents, parents.size()-1, &obj, *render_);
}

void GComputedStyle::print(std::ostream& out) const {
  render_->print(out);
}


GPropMap GComputedStyle::getProps() const {
  GPropMap map;

  if (render_->specif[GPropType::Flow] > 0) {
    map[GPropType::Flow] = *new GFlow(render_->flow_);
  }
  if (render_->specif[GPropType::Flex] > 0) {
    map[GPropType::Flex] = new GFlex(render_->flex_);
  }
  if (render_->specif[GPropType::Align]  > 0) {
    map[GPropType::Align] = new GAlign(render_->align_);
  }
  if (render_->specif[GPropType::Gap]  > 0) {
    map[GPropType::Gap] = new GGap(render_->gap_);
  }
  if (render_->specif[GPropType::Font]  > 0) {
    map[GPropType::Font] = new GFont(render_->font_);
  }
  if (render_->specif[GPropType::Pos]  > 0) {
    map[GPropType::Pos] = new GPos(render_->pos_);
  }
  if (render_->specif[GPropType::Size]  > 0) {
    map[GPropType::Size] = new GSize(render_->size_);
  }

  if (render_->background_ && render_->specif[GPropType::Background]  > 0) {
    map[GPropType::Background] = new GBackground(*render_->background_);
  }
  if (render_->border_ && render_->specif[GPropType::Border]  > 0) {
    map[GPropType::Border] = new GBorder(*render_->border_);
  }
  if (render_->color_ && render_->specif[GPropType::Color]  > 0) {
    map[GPropType::Color] = new GColor(*render_->color_);
  }
  if (render_->format_ && render_->specif[GPropType::Format]  > 0) {
    map[GPropType::Format] = new GFormat(*render_->format_);
  }
  if (render_->icon_ && render_->specif[GPropType::Icon]  > 0) {
    map[GPropType::Icon] = new GIcon(*render_->icon_);
  }
  if (render_->margin_ && render_->specif[GPropType::Margin]  > 0) {
    map[GPropType::Margin] = *new GMargin(*render_->margin_);
  }
  if (render_->text_ && render_->specif[GPropType::Text]  > 0) {
    map[GPropType::Text] = new GText(*render_->text_);
  }
  if (render_->hotkey_ && render_->specif[GPropType::Hotkey]  > 0) {
    map[GPropType::Hotkey] = *new GText(*render_->hotkey_);
  }
  return map;
}

gptr<GProp> GComputedStyle::getProp(GPropType& t) const {
  GProp* prop{};

  switch (t.role()) {
    case GPropType::Text:
      if (render_->text_) prop = new GText(*render_->text_);
      break;

    case GPropType::Icon:
      if (render_->icon_) prop = new GIcon(*render_->icon_);
      break;

    case GPropType::Format:
      if (render_->format_) prop = new GFormat(*render_->format_);
      break;

    case GPropType::Color:
      if (render_->color_) prop = new GColor(*render_->color_);
      break;

    case GPropType::Background:
      if (render_->background_) prop = new GBackground(*render_->background_);
      break;

    case GPropType::Font:
      prop = new GFont(render_->font_);
      break;

    case GPropType::Pos:
      prop = new GPos(render_->pos_);
      break;

    case GPropType::Size:
      prop = new GSize(render_->size_);
      break;

    case GPropType::Flow:
      prop = new GFlow(render_->flow_);
      break;

    case GPropType::Flex:
      prop = new GFlex(render_->flex_);
      break;

    case GPropType::Align:
      prop = new GAlign(render_->align_) ;
      break;

    case GPropType::Gap:
      prop = new GGap(render_->gap_) ;
      break;

    case GPropType::Margin:
      if (render_->margin_) prop = new GMargin(*render_->margin_);
      break;

    case GPropType::Border:
      if (render_->border_) prop = new GBorder(*render_->border_);
      break;
  }

  return prop;
}

}
