//
// << Adding operator: operator<<()
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Add_hpp
#define Guit_Add_hpp
#include <gadget.hpp>
namespace guit {

/// @internal.
class GItemSpec {
public:
  GItemSpec(const char* spec) : spec_(spec) {}
  GItemSpec(const GString& spec) : spec_(spec) {}
  GString spec_;
  mutable std::vector<gptr<GProp>> props_;
};

/// @internal.
class GadgetHolder : public GProp {
public:
  GadgetHolder(Gadget& g) : gadget_(g) {}
  static GPropType& Type();
  GPropType& type() const override {return Type();}
  gptr<Gadget> gadget_;
};

/// Adds or replaces a child or a Prop of a Gadget.
///
/// For **non-box** gadgets (gadgets not deriving from GBox), _child_ can be:
/// - A String, which sets the title or the value of the gadget (e.g. the title of a Window, Box,
///    Label, Button, Item; the value of a TextField, TextArea, Slider, ChartItem)
/// - A Property (GProp), including conditional props (GCond)
/// - A Callback function fired by the on::action trigger. This trigger is gadget dependent and corresponds
///    to the most common action for this gadget (@see gadgetdoc)
/// - A Menu (GMenu) or a Window (GWindow, GDialog, etc.) that will be opened automatically
///    if the gadget is a Button or an Item (GItem) when it is clicked.
///
/// A gadget can have one Prop of a given type, but as many callbacks and  conditional props as needed.
/// Thus, except in the two latter cases, adding a Prop replaces the previous prop of the same type
/// (e.g. if a blue then a red GColor is added to a gadget, its color will be red).
/// These children are added to the *proplist* of the gadget and can be retrieved via the findProp() method.
///
/// In addition, for **box gadgets** (container gadgets deriving from GBox):
/// - _child_ can be any other gadget.
/// -  A String is converted into a GItem that is added to the box. A GItem is a polymorphic Button/Item
///  that adapts to its parent (e.g. it will look as a menu item in a Menu, a list item in a ListBox or
///  a ComboBox, etc.).  As many Strings as needed can be added.
///
/// These children are added to the *childlist* of the box and can be retrieved via the child() method.
///
/// @see Gadget and GBox classes.
/// @note GBox also provides methods for adding gadgets inside the borders of the box.
/// @{
GItemSpec const& operator<<(GItemSpec const& item, GString const& str);
GItemSpec const& operator<<(GItemSpec const& item, GProp& prop);
GItemSpec const& operator<<(GItemSpec const& item, gptr<GProp> prop);

GItemSpec const& operator<<(GItemSpec const& item, GFunction const& fun);
GItemSpec const& operator<<(GItemSpec const& item, GEventFunction const& fun);

GItemSpec const& operator<<(GItemSpec const& item, GMenu& menu);
GItemSpec const& operator<<(GItemSpec const& item, gptr<GMenu> menu);

GItemSpec const& operator<<(GItemSpec const& item, GWindow& win);
GItemSpec const& operator<<(GItemSpec const& item, gptr<GWindow> win);

template <class T, class = typename T::IsGadget>
inline T& operator<<(T& gadget, const char* str) {
  if (str) gadget.add(str);
  return gadget;
}

template <class T, class = typename T::IsGadget>
inline T& operator<<(T& gadget, GString const& str) {
  gadget.add(str);
  return gadget;
}

template <class T, class = typename T::IsGadget>
inline T& operator<<(T& gadget, GStrings const& str) {
  gadget.add(str);
  return gadget;
}

template <class T, class = typename T::IsGadget>
inline T& operator<<(T& gadget, GItemSpec const& item) {
  gadget.addItem(item);
  return gadget;
}

template <class T, class Child, class = typename T::IsGadget, class = typename Child::IsGadgetChild>
inline T& operator<<(T& gadget, Child& child) {
  gadget.add(child);
  return gadget;
}

template <class T, class Child, class = typename T::IsGadget, class = typename Child::IsGadgetChild>
inline T& operator<<(T& gadget, Child* child) {
  gadget.add(child);
  return gadget;
}

template <class T, class Child, class = typename T::IsGadget>
inline T& operator<<(T& gadget, gptr<Child> child) {
  gadget.add(child.get());
  return gadget;
}

template <class T, class = typename T::IsGadgetOrVarProp>
inline T& operator<<(T& parent, GFunction const& fun) {
  parent.addFun(fun);
  return parent;
}

template <class T, class = typename T::IsGadgetOrVarProp>
inline T& operator<<(T& parent, GEventFunction const& fun) {
  parent.addFun(fun);
  return parent;
}

template <class T, class = typename T::IsGadgetOrVarProp>
inline T& operator<<(T& parent, GCaller& fun) {
  parent.addFun(fun);
  return parent;
}

template <class T, class Child, class = typename T::IsGadget>
inline T& operator<<(T& gadget, Child&(creator)(GString const&)) {
  gadget.add((creator)(""));
  return gadget;
}

template <class T, class = typename T::IsGadget>
inline T& operator<<(T& (creator)(GString const&), GString const& str) {
  return (creator)("") << str;
}

template <class T, class = typename T::IsGadget>
inline T& operator<<(T& (creator)(GString const&), GItemSpec const& item) {
  return (creator)("") << item;
}

template <class T, class Child, class = typename T::IsGadget, class = typename Child::IsGadgetChild>
inline T& operator<<(T& (creator)(GString const&), Child& child) {
  return (creator)("") << child;
}

template <class T, class Child, class = typename T::IsGadget>
inline T& operator<<(T& (creator)(GString const&), gptr<Child> child) {
  return (creator)("") << child;
}

template <class T, class = typename T::IsGadget>
inline T& operator<<(T& (creator)(GString const&), GFunction const& fun) {
  return (creator)("") << fun;
}

template <class T, class = typename T::IsGadget>
inline T& operator<<(T& (creator)(GString const&), GEventFunction const& fun) {
  return (creator)("") << fun;
}
/// @}

}
#endif
