//
//  Boxes that enforce mutual exclusion of their items.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_ChoiceBox_hpp
#define Guit_ChoiceBox_hpp
#include <gbox.hpp>
#include <gchoice.hpp>
namespace guit {

/// Box that enforce mutual exclusion of its children.
/// This class is the base class of boxes that enforce mutual exclusion among their children
/// (eg GListBox, GRadioBox, GStackBox, GTabBox, GToolBox, GOptionBox, GComboBox, GChart).
/// Technically, a GChoiceBox is a GBox with a GChoice prop and methods that help selecting children.
/// Anyway, the GChoice implements exclusive selection, thus any GBox can become exclusive just
/// by adding a GChoice prop to it.
///
/// The choose() methods select a child, the chosenItem(), chosenIndex(), chosenText() retrieve, repectively,
/// the selected child, the selection index, and the text of the selected child. The same operations
/// can be performed through the associated GChoice prop, which is returned by the choice() method.
///
/// Only GItem (and derived) children are by default "checkable" (ie selectable), but any child gadget
/// can be made selectable by calling its setCheckable() method (or by adding a GCkeck prop to it).
/// Importantly, the selection index takes only into account these children (other children are ignored).
///
/// Any gadget can be added to a GChoiceBox, but they generally contain GItem children (or subclasses).
/// Adding a string to a GChoiceBox automatically converts it into an instance of a GItem subclass.
/// The actual class depends on the box, e.g. GChoiceItem for a GChoiceBox, GRadioItem for a
/// GRadioBox, GChartItem for a GChartBox. Multiple items can be created and added as folllows:
/// @code
///    auto& mycbox = ChoiceBox()     // equivalent to *new GChoiceBox()
///    << Apples" << "Oranges" << "Cherries";
///
///    // GStrings is an alias for std::vector<std::string>
///    GStrings items{Apples", "Oranges", "Cherries"};
///
///    auto& mylist = ListBox() << items;
/// @endcode
/// They can also be created using CSS styles (both solutions can be used simultaneoulsy):
/// @code
///    auto& mychart = ChartBox("#mychart");
///
///    // CSS
///    #mychart {items: Apples, Oranges, Cherries}
/// @endcode
///
/// The default orientation is GFlow::column (vertical). It can be made horizontal through CSS styling
/// or by adding a GFlow::row prop. Similarly, GFlow::pie makes the box to be displayed as a pie:
////// @code
///    mylist << GFlow::row;
///
///    // CSS
///    #mychart {flow:pie; items: Apples, Oranges, Cherries}
/// @endcode
///
/// Multiple choice boxes can be synced by making them to share the same GChoice prop:
/// @code
///    auto& choice = Choice();
///    mycbox << choice;
///    mylist << choice;
///    mychart << choice;
/// @endcode
/// Calling choose() on any of these boxes then selects all of them accordingly. They can also be
/// selected by setting choose directly:
/// @code
///    choice = 1;     // selects "Oranges" (as the index starts at 0)
///    std::cerr << choice() << std::endl;    // prints 1
/// @endcode
/// The GChoice prop can be bound with another prop using the bind() function or the  <<= operator,
///
class GChoiceBox : public GBox {
protected:
  GChoiceBox(GadgetType&, GString const& css_and_title);

public:
  ~GChoiceBox();
  GChoiceBox(GString const& css_and_title = "");
  GChoiceBox& clone(GClone const&) override;
  
  GChoiceBox* toChoiceBox() override {return this;}

  /// Selects (ie checks) a child.
  /// - choose(item): item must be one of the box' children; uncchecks all if item is null or N/A
  /// - choose(index): index is the position in the box' child list (no taking into account children that
  ///  are not checkable: see Gadget::isCheckable()); deselects all if index is negative or N/A
  /// - choose(string): selects the fist child whose text() matches with 'string'; deselects all if no match
  /// @{
  virtual void choose(Gadget* item);
  virtual void choose(Gadget& item) {GChoiceBox::choose(&item);}
  virtual void choose(int index);
  virtual void choose(GString const& string);
  /// @}

  /// Returns the checked child, its index or its text.
  /// - chosen() returns the checked child, or null if none selected
  /// - chosenIndex() returns the index of the checked child in the child list , taking only into account
  ///  checkable children (see Gadget::setCheckable()), or -1 if none selected
  /// - chosenText() returns the text() of the checked child, or the empty string if none selected
  /// @{
  virtual Gadget* chosen() const;
  virtual int     chosenIndex() const;
  virtual GString chosenText() const;
  /// @}

  /// Sets/returns the value.
  /// The value of a GChoiceBox is the index of the selected item. Thus, setValue() is equivalent to
  /// choose() and value() to chosenIndex(). stringValue() returns the index conversted into a string.
  /// @{
  void setValue(GString const& string) override {choose(string);}
  int  value() const;
  GString stringValue() const override;
  /// @}

  /// Returns the prop that controls the selection.
  /// @{
  GChoice& choice();
  GChoice* choiceProp() override {return choice_;}
  GProp*   valueProp() override {return choice_;}
  /// @}

  bool hasValue() const override {return true;}

  using GType = GadgetType_<GChoiceBox,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// Guit provides full access to all fields, but beware using them is at your own risk.
  /// @{
  gptr<GChoice> choice_;
  void initClone(Gadget& from, GClone const&) override;
  void addProp(GProp&, bool add_to_proplist) override;
  void setChoiceImpl(GChoice& c, bool add_to_list) override;
  virtual GBox& itemBox() {return *this;}
  virtual GBox const& itemBox() const {return *this;}
  Gadget& createDefaultItem(GString const& arg) override;
  /// @}
};

GChoiceBox& ChoiceBox(GString const& css_and_title = "");



/// List of exclusive items.
/// Works as GChoiceBox but with a different look and feel
class GListBox : public GChoiceBox {
protected:
  GListBox(GadgetType&, GString const& css_and_title);

public:
  GListBox(GString const& css_and_title = "");
  GListBox& clone(GClone const&) override;

  using GType = GadgetType_<GListBox,GChoiceBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};

GListBox& ListBox(GString const& css_and_title = "");



/// Group of exclusive radio items.
class GRadioBox : public GChoiceBox {
protected:
  GRadioBox(GadgetType&, GString const& css_and_title);

public:
  GRadioBox(GString const& css_and_title = "");
  GRadioBox& clone(GClone const&) override;

  using GType = GadgetType_<GRadioBox,GChoiceBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};

GRadioBox& RadioBox(GString const& css_and_title = "");

}
#endif
