//
//  Gadgets that open a menu of options
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gevent.hpp>
#include <gtrigger.hpp>
#include <gcombobox.hpp>
#include <gtextitem.hpp>
#include <gicon.hpp>
#include <gmenu.hpp>
#include <gcontrol.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(OptionBox)
GIMPLEMENT_GADGET(ComboBox)

GOptionBox::~GOptionBox() = default;

GOptionBox::GOptionBox(GadgetType& t, GString const& spec) : GChoiceBox(t, spec),
menu_(new GMenu())
{
  // remove choice_ otherwise cliking on the item will reset the selection
  remove(choice_);

  // choice_ changes the selected item in the menu
  menu_->modes_.choiceBehavior = uint8_t(ChoiceBehavior::Select);

  // order matters: changeCB must not be called before adding choice_
  *menu_ << choice_;
  *menu_ << on::change / [this](GEvent& e){changeCB(e);};
}

GOptionBox::GOptionBox(GString const& spec) : GOptionBox(Type(), spec) {
  selector_ = *new GItem << GIcon::down << *menu_;
  GBox::add(selector_, 0);
}

GOptionBox& GOptionBox::clone(GClone const& c) {
  auto* g = new GOptionBox;
  /*   done initClone()
  if (c & GClone::ShareProps) {
    box->choice_ = choice_;
    *box->menu_ << choice_;
  }*/
  g->initClone(*this, c);
  return *g;
}

void GOptionBox::initClone(Gadget& from, GClone const& c) {
  // as the item are cloned they must not be created from styles
  settings_.itemsSet = 1;
  // Gadget::initClone pas GBox::initClone car sinon on va recopier
  // le bouton et son contenu
  Gadget::initClone(from, c);

  if (c & GClone::ShareProps) {
    if (auto* f = from.toChoiceBox()) {
      choice_ = f->choice_;
      *menu_ << f->choice_;
    }
  }

  if (c & GClone::Children) {
    if (auto* f = dynamic_cast<GOptionBox*>(&from)) {
      menu_->initClone(*f->menu_, c);
    }
  }
}

void GOptionBox::init(GRender* parrender) {
  if (omodes_.initialized) return;
  GChoiceBox::init(parrender);
  // needed to init the selector
  if (*choice_ >= 0) {
    if (auto* item = choice_->chosenItem(*menu_)) *selector_ << item->text();
  }
}

void GOptionBox::changeCB(GEvent& e) {
  Gadget* g = e.chosen();
  if (g) *selector_ << g->text();
  else *selector_ << " ";
  // propager childchange a la box, pas action qui doit etre fait sur un clic
  doFire(e);
}

void GOptionBox::initChildren(GRender* parrender) {
  GBox::initChildren(parrender);
  menu_->init(parrender);
}

void GOptionBox::setChoiceImpl(GChoice& c, bool add_to_list) {
  //modes_.hasChoice = 1;
  choice_ = c;
  menu_ << c;
}



GBox& GOptionBox::itemBox() {return *menu_;}

GBox const& GOptionBox::itemBox() const {return *menu_;}

GItem& GOptionBox::selector() {return *selector_;}

GMenu& GOptionBox::menu() {return *menu_;}

void GOptionBox::add(Gadget* child, int ix) {
  menu_->add(child, ix);
}

bool GOptionBox::remove(Gadget* child) {
  return menu_->remove(child);
}

bool GOptionBox::removeAt(int index)  {
  return menu_->removeAt(index);
}

void GOptionBox::removeChildren(Remove mode) {
  menu_->removeChildren(mode);
}

// - - - - - - - 

GComboBox::GComboBox(GadgetType& t, GString const& spec) : GOptionBox(t, spec) {
  auto&txt = *new GTextField << GIcon::down;

  selector_ = txt;

  //*menu_ << txt.value();

  GBox::add(selector_, 0);

  // the menu is added to the box, not to the texfield otherwise clicking on
  // the textfield would open the menu
  GBox::add(menu_, 1);

  *selector_
  << on::mousedown / [this](GEvent& e){
    auto& ee = e.asMouseEvent();
    if (ee.x() > w() - GPrefs::iconWidth()) menu_->show(ee, true);
  }
  << on::mouseup / [this]{menu_->show(false);};
}

GComboBox::GComboBox(GString const& spec)
: GComboBox(Type(), spec) {}

GComboBox& GComboBox::clone(GClone const& c) {
  auto* g = new GComboBox;
  /*
  if (c & GClone::ShareProps) {
    box->choice_ = choice_;
    *box->menu_ << choice_;      // should also clone textfield value?
  }*/
  g->initClone(*this, c);
  return *g;
}

GTextItem& GComboBox::selector() {
  return *selector_->toTextItem();
}

}
