//
// File Dialog
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

// for Windows
#define _CRT_SECURE_NO_WARNINGS 1

#include <chrono>
#include <filesystem>
#include <gapp.hpp>
#include <gtrigger.hpp>
#include <gevent.hpp>
#include <gfile.hpp>
#include <gfiledialog.hpp>
#include <gscrollbox.hpp>
#include <gexpr.hpp>
#include <gborder.hpp>
#include <gcolor.hpp>
#include <gsize.hpp>
#include <gmodes.hpp>
#include <gtextitem.hpp>
#include <gchoicebox.hpp>
#include <gtable.hpp>
#include <gcontrol.hpp>
#include <gstyle.hpp>
using namespace std;
namespace fs = std::filesystem;
//using namespace std::chrono;
namespace guit {

static const size_t MaxFileNameSize = 50;           // TODO!!!

GIMPLEMENT_GADGET(FileDialog)

GFileDialog::GFileDialog(GString const& spec) :
GFileDialog(Type(), spec) {}

GFileDialog::GFileDialog(GadgetType& t, GString const& spec) : GDialog(t, spec),
choice_(new GChoice(-1)),
list_(*new GTable(":FileTable") << *choice_),
file_(TextField()),
dir_(TextField())
{
  setModal(true);
  *this
  << (Table("{margin:25 20 0 5; flex:1 0}")    // should be in styles !!
      << (TRow
          << "{align:e;font:bold}File:"s
          << *file_
          << " "
          << (LinkItem("{font:small}Parent")
              << [this]{setDir("..");}
              )
          << (LinkItem("{font:small}Home")
              << [this]{setDir("~");}
              )
          )
      << (TRow
          << "{align:e;font:bold}Directory:"s
          << (*dir_ << [this]{setDir(dir_->text());})
          << " "
          << (LinkItem("{font:small}Desktop")
              // << [this]{setDir(GApp::desktopDir());}
              << [this]{setDir("~/Desktop");}
              )
          << (LinkItem("{font:small}Resources")
              << [this]{setDir(":");}
              )
          // << (Button("<b>.") << [this]{setDir(".");})   // launch dir
          << "  "
          )
      )
  << (ScrollBox << GBackground::white << *list_);
  
  setButtons("Cancel,OK");
}

GFileDialog& GFileDialog::clone(GClone const& cm) {
  auto* g = new GFileDialog;
  g->initClone(*this, cm);
  return *g;
}

GString GFileDialog::pathname() const {
  return dir_->text() + '/' + file_->text();
}

GString GFileDialog::basename() const {
  return file_->text();
}

GString GFileDialog::dirname() const {
  return dir_->text();
}

GFileDialog& GFileDialog::setDir(GString const& path) {
  //cerr << "GFileDialog::setDir: "<< path << endl;
  GString fullpath;
  if (path == "..") {
    fullpath = parentPath(dir_->text());
  }
  /*
  else if (path == ".") {
    fullpath = getenvvar("PWD");
  }*/
  else {
    fullpath = normPath(path, false);
  }
  if (fullpath.empty()) fullpath = "/";
  
  *dir_ << fullpath;
  rescan();
  return *this;
}

template <typename TP> std::time_t to_time_t(TP tp) {
  using namespace std::chrono;
  auto sctp = time_point_cast<system_clock::duration>(tp - TP::clock::now()
                                                      + system_clock::now());
  return system_clock::to_time_t(sctp);
}


void GFileDialog::rescan() {
  *file_ << "";
  list_->removeChildren();
  auto dirpath = normPath(dir_->text());
  if (dirpath.empty()) return;
  //cerr << "DirPath: " << dirpath << endl;

  try {
    for (auto& p : fs::directory_iterator(dirpath)) {
      //cerr << p.path() << '\n';

      GString fname = p.path().filename().string();
      if (fname.size() > MaxFileNameSize) {
        fname.erase(MaxFileNameSize);
        fname += "...";
      }

      bool isdir{};
      GString icon;

       if (p.is_directory()) {icon = ">"; isdir = true;}
       else if (!p.is_regular_file()) icon = "x";

      //else if (!p.executable()) icon = "x";

      std::error_code ec1{};
      GString size = isdir ? "--" : to_string(p.file_size(ec1));

      std::error_code ec2{};
      auto t = p.last_write_time(ec2);
      //auto tt = decltype(t)::clock::to_time_t(t);  // not yet multiplaform
      auto tt = to_time_t(t);
      GString time = std::asctime(std::localtime(&tt));
      if (!time.empty() && time.back() == '\n') time.pop_back();

      auto& sel = *new GCheck(false);
      auto& cb = on::dblclick / [this,isdir]{dblClick(isdir);};

      *list_
      << (*new GTRow(":FileRow")
          << sel
          << on::checked / [this,fname]{selectEntry(fname);}
          << (*new GItem(":FileItem {margin:8 8 0 0}") << sel << cb << icon)
          << (*new GItem(":FileItem") << fname << sel << cb)
          << (*new GItem(":FileItem") << time << sel << cb)
          << (*new GItem(":FileItem {flex:0 0}") << size << sel << cb)
          );
    }
  }
  catch (fs::filesystem_error&) {}
}

void GFileDialog::selectEntry(GString fname) {
  *file_ = fname;
}

void GFileDialog::dblClick(bool isdir) {
  //cerr << "dblClick: " << fileName() << endl;
  if (isdir) setDir(pathname()); else setResult(1, true);
}

GFileDialog& GFileDialog::alertIf(GBool& test) {
  alertif_ = &test;
  return *this;
}

GFileDialog& GFileDialog::alertIf(GExprBool const& test) {
  alertif_ = &test.prop;
  return *this;
}

GFileDialog& GFileDialog::alertIf(std::function<bool()> test) {
  alertif_ = &GExprBool(test).prop;
  return *this;
}

GFileDialog& GFileDialog::alertFileExist(bool state) {
  alert_file_exist_ = state;
  return *this;
}

GFileDialog& GFileDialog::alertFileDontExist(bool state) {
  alert_file_dont_exist_ = state;
  return *this;
}

GFileDialog& GFileDialog::show(bool state) {
  if (!state) {
    GDialog::show(false);
    return *this;
  }
  if (alertif_
      && alertif_->boolValue()
      && !GDialog::confirm("Do you want to open a file?", "Changes will be lost!",
                           "Cancel,Proceed")) {
    return *this;
  }
  /*
  if (dir_->text().empty()) {
    //const char* s = ::getenv("PWD");
    //if (s) *dir_ << s;
    //else *dir_ = std::string(1, GPathSepar);
    *dir_ << getenvvar("PWD");
  }*/
  rescan();
  GDialog::show(true);
  return *this;
}

void GFileDialog::setResult(int result, bool hide) {
  if (result == 1) {
    if (basename().empty()) return;

    /* pbm: empeche de selectionner un directory!
     // change de directory qd on clique sur open
     if (fileinfo_ && fileinfo_->isDir()) {   // wrong if File string edited!
     setDir(fileName());
     return;
    }*/

    if (alert_file_exist_) {
      if (fileExist(pathname())
          && !GDialog::confirm("This file already exists","Do you want to replace it?",
                               "Cancel,Replace")) {
        return;
      }
    }
    else if (alert_file_dont_exist_) {
      if (!fileExist(pathname())
          && !GDialog::confirm("This file don't exist","","Cancel,Continue")) {
        return;
      }
    }
  }

  result_ = result;
  quitloop_ = hide;
  if (result == 1) {
    GEvent e(GTrigger::action, GCore::elapsedTime());
    e.gadget_ = this;
    //GTrigger::setCurrentEvent(&e);
    props_.fire(e);
    //GTrigger::setCurrentEvent(nullptr);
  }
  if (hide) close();
}


GOpenDialog::GOpenDialog(GString const& css_and_title)
: GFileDialog(Type(), css_and_title) {}

GSaveDialog::GSaveDialog(GString const& css_and_title)
: GFileDialog(Type(), css_and_title) {}


GOpenDialog::GType& GOpenDialog::Type() {
  static auto& t = *new GOpenDialog::GType("OpenDialog");
  return t;
}

GOpenDialog& OpenDialog(GString const& title, GString const& dir) {
  auto& d = *new GOpenDialog(title.empty() ? "Open" : title);
  d.setButtons("Cancel,Open");
  d.setDir(dir.empty() ? GPrefs::openDir() : dir);
  d.alertFileDontExist(true);
  return d;
}

GOpenDialog& GOpenDialog::clone(GClone const& cm) {
  auto* g = new GOpenDialog();
  g->initClone(*this, cm);
  return *g;
}

GSaveDialog::GType& GSaveDialog::Type() {
  static auto& t = *new GSaveDialog::GType("SaveDialog");
  return t;
}

GSaveDialog& SaveDialog(GString const& title, GString const& dir) {
  auto& d = *new GSaveDialog(title.empty() ? "Save" : title);
  d.setButtons("Cancel,Save");
  d.setDir(dir.empty() ? GPrefs::saveDir() : dir);
  d.alertFileExist(true);
  return d;
}

GSaveDialog& GSaveDialog::clone(GClone const& cm) {
  auto* g = new GSaveDialog();
  g->initClone(*this, cm);
  return *g;
}

}
