//
//  Item Gadgets with a numerical value
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_ValueItem_hpp
#define Guit_ValueItem_hpp
#include <gitem.hpp>
#include <gfloat.hpp>
namespace guit {

/// Item gadgets with a numerical value.
/// Base for GChartItem, GSlider, GScrollBar.
/// - A GNumber prop (e.g. GFloat or GInt) can be added to this gadget to control its value.
/// - A GText prop can be added to this gadget to control the text it displays. The color of the
///  changes automatically depending on the GBackground color of the gadget (white if the
///  background color is dark, black otherwise)
class GValueItem : public GItem {
protected:
  GValueItem(GadgetType& c) : GItem(c) {}
  GValueItem(GadgetType&, GString const& css_value_title);

public:
  ~GValueItem();
  explicit GValueItem(GString const& css_value_title = "");
  explicit GValueItem(float value);
  GValueItem& clone(GClone const&) override;

  GValueItem* toValueItem() override {return this;}

  /// Returns the current value.
  /// Returns the value of the associated GNumber prop (see operator()).
  /// The value lies between min() and max() (0 and 100 by default)
  /// @{
  float value() const;
  GString stringValue() const override;
  /// @}

  /// Changes the current value.
  /// Changes the value of the associated GNumber prop (see operator()).
  /// _value_ should lie between min() and max(), otherwise the gadget's value is set to
  /// either min() or max() (0 and 100 by default)
  /// @{
  virtual void setValue(float value);
  void setValue(GString const&) override;
  /// @}

  /// Returns/changes the minimum and maximum values.
  /// Returns/changes the min and max value of the associated GNumber prop (see operator()).
  /// The default min is 0, the default max is 100.
  /// @{
  GValueItem& setRange(float min, float max);
  float min() const;
  float max() const;
  /// @}

  /// Returns the prop that controls the value.
  /// The prop that controls the value can be changed by using the << operator:
  /// @code
  ///    auto& value = Float();
  ///    auto& item = ValueItem() << value;
  ///    value = 10.f;
  /// @endcode
  /// @{
  GNumber& operator()();
  GProp* valueProp() override {return value_;}
  /// @}
  
  bool hasValue() const override {return true;}

  using GType = GadgetType_<GValueItem,GItem>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  virtual GFlow::Kind getFlow();
  void initSpec(GString const& spec);
  void initClone(Gadget& from, GClone const&) override;
  void addProp(GProp&, bool add_to_proplist) override;
  void doPaint(GRenderGraphics&, GRender* parr) override;
  void movedResizedCB(GRect const& initialframe) override;
  virtual void paintContent(GRenderGraphics&, GRender&, GFlow::Kind dir);
  virtual void mouseDrag(GMouseEvent&);
  //gptr<GNumber> value_;
  gptr<class GFloat> value_;
  float offset_{};
  /// @}
};


/// Creates a GValueItem (Item Gadgets with a numerical value).
/// @{
GValueItem& ValueItem(GString const& css_and_value = "");
GValueItem& ValueItem(float value);
/// @}

}
#endif
