//
// Base class of Guit Objects.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Object_hpp
#define Guit_Object_hpp
#include <gtypes.hpp>
namespace guit {

class Gadget;
class GBox;
class GProp;

/// Base class of Guit Objects.
/// A GObject can be pointed by one or several gptr<> smart pointers. These pointers take ownership,
/// so that the GObject is auto-deleted (*) when the last gptr pointing to it is unset or destroyed..
///
/// Gadget(s) point to their props and children through gptrs, thus props and children are auto-deleted (*)
/// when removed from the last parent containing them, if not pointed by another gptr. gptrs should
/// thus be used when removing objects from their parents if these objects need to be reused.
///
/// (*) Objects created in the stack or in static memory are not auto-deleted by gptrs. gptrs can thus
/// point objects allocated in the heap, the stack or in static memory.
///
/// Raw C++ pointers and references should only be used as temporary references (e.g. when
/// creating a gadget tree) as they may become pending at some point.
///
/// std::shared_ptr, std::weak_ptr, std::unique_ptr CANNOT BE USED.
///
/// The delete operator does nothing except issuing a warning. An object pointed by a raw pointer
/// can be deleted as follows (this will delete it only if it is not gptr-referenced elsewhere):
/// @code
///    MyClass* p = new MyClass();
///    // etc.
///    gptr<GObject> p2 = p;
///    p2.reset();      // or p2 = nullptr;
/// @endcode
///
/// Auto-deletion can be temporarily or permanently disallowed by calling (respectively)
/// forgetSmartPointers() and ignoreSmartPointers().
/// @see gptr.
class GObject {
public:
  using IsGObject = bool;

  GObject();
  virtual ~GObject() = default;

  virtual GString typeName() const = 0;

  /// Prints an error message.
  /// @see GApp::onError() to change error management.
  virtual void error(GString const& funname, GString const& message) const;

  /// Class conversions.
  /// @{
  template <class Subclass>
  inline Subclass* to() {return dynamic_cast<Subclass*>(this);}

  template <class Subclass>
  inline Subclass const* to() const {return dynamic_cast<Subclass const*>(this);}

  virtual Gadget* toGadget() {return nullptr;}
  virtual GBox*   toBox()    {return nullptr;}
  virtual GProp*  toProp()   {return nullptr;}

  virtual Gadget const* toGadget() const {return nullptr;}
  virtual GBox   const* toBox()    const {return nullptr;}
  virtual GProp  const* toProp()   const {return nullptr;}
  /// @}

  /// delete does nothing, see GObject class.
  void operator delete(void*);

  void* operator new(std::size_t);

  /// Checks/sets whether this object can be auto-deleted by smart pointers.
  /// ignoreSmartPointers() has a permanent effect, contrary to forgetSmartPointers().
  /// Objects not created by calling new are never auto-deleted by smart pointers.
  /// @{
  void ignoreSmartPointers() {omodes_.plain = 1;}
  bool isIgnoringSmartPointers() const {return omodes_.plain;}
  /// @}

  /// Forgets all smart pointers pointing this object.
  /// The smart pointers that are currently pointing to the object won't auto-delete it.
  /// forgetSmartPointers() has a temporary effect, contrary to ignoreSmartPointers().
  void forgetSmartPointers() {omodes_.autodel = 0;}

  /// Returns the numbers of smart pointers referencing the object.
  unsigned int useCount() const {return refcount_;}
  
  /// [Implementation] Notifies this object that it must no longer refer _obj_.
  virtual void removeNotifiers(void* obj) {}
  
  // can't trivially copy object.
  GObject(GObject const&) = delete;
  GObject(GObject const&&) = delete;
  GObject& operator=(GObject const&) = delete;
  GObject& operator=(GObject const&&) = delete;
  friend void gaddPtr(GObject*, bool will_autodel);
  friend void gremovePtr(GObject*);

  /// @internal
  /// Guit provides access to almost all fields, but beware using them is at your own risk.
  /// @{
  uint16_t refcount_;
  struct {
    uint8_t plain:1, autodel:1, immutable:1, initialized:1;
  } omodes_;
  /// @}
};

}
#endif
