//
// Prop Binding
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Bind_hpp
#define Guit_Bind_hpp
#include <gvarprop.hpp>
#include <gnotifier.hpp>
#include <gwrapper.hpp>
namespace guit {

/// Unbinds two varprops.
/// Unbinds prop1 from prop2. If the two props were synced, they are unbound both ways.
/// @see gbind()
inline void gunbind(GVarProp& prop1, GVarProp& prop2) {prop1.unbind(prop2);}

/// Unbinds a varprop from all varprops previously bound with this tag.
/// - If _tag_ is a positive value, unbinds all varprops that were bound to _prop_ with this _tag_
///  (whatever they were bound by gbind(prop, other, tag) or gbind(other, prop, tag))
/// - If _tag_ is a negative value, unbinds all varprops that were bound to _prop_ with a tag >= _tag_
/// @{
inline void gunbind(GVarProp& prop, GVarProp::BindTag tag = 0) {prop.unbind(tag);}
/// @}


/// Binds a varprop to another varprop.
/// Binds _prop1_ to _prop2_. prop1 will then be set to the value of prop2 whenever it changes:
/// @code
///   auto& txt = Text();
///   auto& num = Float();
///   gbind(txt, num);   // txt then depends on num
///   // or:
///   txt <<= num;       // operator<<=() has the same effect as gbind()
/// @endcode
/// Props can also be bound to Active Expressions (GExpr) to perform calculations:
/// @code
///    auto& euros = Float(100);
///    auto& dollars = Float();
///    auto& rate = Float(1.2);      // Euro to Dollar conversion rate
///
///    dollars <<= ~euros * ~rate;   // dollars depends on euros
///    euros <<= ~dollars / ~rate;   // euros depends on dollars
/// @endcode
/// - The props can be bound both ways using gsync(), and unbound using gunbind().
/// - The optional _tag_ argument makes it possible to unbind selectively using gunbind()
/// @see GVarProp, gunbind(), gsync() and GExpr (Active Expressions).
/// @{
template <class R, class RM = typename R::IsVarProp, class S, class SM = typename S::IsVarProp>
inline R& gbind(R& prop1, S& prop2, GVarProp::BindTag tag = 0) {
  prop1.bindImpl(prop2, *new GChangePropNotifier<S,R>(prop2,prop1), tag, false);
  return prop1;
}

template <class R, class RM = typename R::IsVarProp, class S, class SM = typename S::IsVarProp>
inline R& operator<<=(R& prop1, S& prop2) {return gbind(prop1, prop2);}

template <class R, class RM = typename R::IsVarProp, class S>
inline R& operator<<=(R& prop1, GPropRef_<S> prop2) {return gbind(prop1, (S&)prop2.prop_);}
/// @}


/// Binds a varprop to a varprop's field, or vice versa.
/// Binds the left hand side to the right hand side. Works as gbind() or operator<<=() except that
/// either the left or right hand side is a varprop's field.
/// @code
///    auto& f = Float();
///    auto& c = Color();
///    f <<= c.redField();   // f depends on the red field of c
///    c.redField() <<= f;   // the red field of c depends on f
/// @endcode
/// @see gbind() and gsync()
/// @{
template <class R, class S, class F, class MATCH = typename R::IsVarProp>
inline R& gbind(R& prop, GPropField<S,F> const& field) {
  using RVal = typename R::ValueType;
  prop.bindImpl(field.prop_,
                *new GChangeFieldNotifier<S,F,R,RVal>(field, GPropField<R,RVal>(prop,prop.value_)),
                0, false);
  return prop;
}

template <class R, class F, class S>
inline R& gbind(GPropField<R,F> const& field, GVarProp_<S>& prop) {
  field.prop_.bindImpl(prop, *new GChangeFieldNotifier<GVarProp_<S>,S,R,F>
                       (GPropField<GVarProp_<S>,F>(prop,prop.value_), field), 0, false);
  return field.prop_;
}

template <class R, class S, class F, class MATCH = typename R::IsVarProp>
inline R& operator<<=(R& prop, GPropField<S,F> const& field) {return gbind(prop,field);}

template <class R, class F, class S>
inline R& operator<<=(GPropField<R,F> const& field, GVarProp_<S>& prop) {return gbind(field,prop);}
/// @}


/// Binds a varprop's field to another varprop's field.
/// Binds the left hand side to the right hand side. Works as gbind() or operator<<=() except that
/// both sides are a varprop's field.
/// @code
///    auto& c1 = Color();
///    auto& c2 = Color();
///    c1.redField() <<= c2.blueField();   // c1 red field depends on c2 blue field
/// @endcode
/// @see gbind() and gsync()
/// @{
template <class R, class RF, class S, class SF>
inline R& gbind(GPropField<R,RF> const& field1, GPropField<S,SF> const& field2) {
  field1.prop_.bindImpl(field2.prop_,
                        *new GChangeFieldNotifier<S,SF,R,RF>(field2,field1),
                        0, false);
  return field1.prop_;
}

template <class R, class RF, class S, class SF>
inline R& operator<<=(GPropField<R,RF> const& field1, GPropField<S,SF> const& field2) {
  return gbind(field1, field2);
}
/// @}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Syncs two varprops.
/// Binds the props both ways and enforces constraints (see below)
/// @code
///   auto& txt = Text();
///   auto& num = Float();
///   gsync(txt, num);   // the 2props are synced
///   // or:
///   txt ^= num;        // operator^=() has the same effect as gsync()
/// @endcode
/// The right-hand prop (_num_) imposes its constraints to the left-hand prop (_txt_): _txt_ will be forced
/// to contain a string convertible to a float. If _txt_ is  added to a GTextField or GTextArea the user
/// will be forced to enter a (floating) number.
/// @see gbind() for details.
/// @{
template <class R, class RM = typename R::IsVarProp,
class S, class SM = typename S::IsVarProp>
inline R& gsync(R& prop1, S& prop2, GVarProp::BindTag tag = 0) {
  prop1.bindImpl(prop2, *new GChangePropNotifier<S,R>(prop2,prop1), tag, false);
  prop2.bindImpl(prop1, *new GChangePropNotifier<R,S>(prop1,prop2), tag, true);
  return prop1;
}
template <class R, class RM = typename R::IsVarProp,
class S, class SM = typename S::IsVarProp>
inline R& operator^=(R& prop1, S& prop2) {return gsync(prop1, prop2);}
/// @}


/// Syncs a varprop with a field of another varprop, or vice-versa.
/// Syncs the left hand side with the right hand side. Works as gsync() or operator^=() except that
/// either the left or right hand side is a varprop's field.
/// @code
///    auto& f = Float();
///    auto& c = Color();
///    f ^= c.redField();   // f and the red field of c are synced
/// @endcode
/// @see gsync().
/// @{
template <class R, class S, class F, class MATCH = typename R::IsVarProp>
inline R& gsync(R& prop, GPropField<S,F> const& field) {
  using RVal = typename R::ValueType;
  gbind<R,S,F>(prop, field);
  field.prop_.bindImpl(prop, *new GChangeFieldNotifier<R,RVal,S,F>
                       (GPropField<R,RVal>(prop,prop.value_), field), 0, false);
  return prop;
}

template <class R, class F, class S>
inline R& gsync(GPropField<R,F> const& field, GVarProp_<S>& prop) {
  gbind(field, prop);
  prop.bindImpl(field.prop_, *new GChangeFieldNotifier<R,F,GVarProp_<S>,S>
                (field, GPropField<GVarProp_<S>,S>(prop,prop.value_)), 0, false);
  return field.prop_;
}

template <class R, class S, class F, class MATCH = typename R::IsVarProp>
inline R& operator^=(R& prop, GPropField<S,F> const& field) {
  return gsync<R,S,F>(prop, field);
}

template <class R, class F, class S>
inline R& operator^=(GPropField<R,F> const& field, GVarProp_<S>& prop) {
  return gsync<R,F,S>(field,prop);
}
/// @}


/// Syncs a varprop's field with another varprop's field.
/// Syncs the left hand side with the right hand side. Works as gsync() or operator^=() except that
/// both sides are varprop's fields.
/// @code
///    auto& c1 = Color();
///    auto& c2 = Color();
///    c1.redField() ^= c2.blueField();    // both fields are synced
/// @endcode
/// @see gbind().
/// @{
template <class R, class RF, class S, class SF>
inline R& gsync(GPropField<R,RF> const& field1, GPropField<S,SF> const& field2) {
  gbind(field1, field2);
  gbind(field2, field1);
  return field1.prop_;
}

template <class R, class RF, class S, class SF>
inline R& operator^=(GPropField<R,RF> const& field1, GPropField<S,SF> const& field2) {
  return gsync(field1, field2);
}
/// @}

}
#endif
