//
// Bool Prop
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitBool_hpp
#define GuitBool_hpp
#include <gvarprop.hpp>
namespace guit {

/// Boolean prop.
class GBool : public GVarProp_<bool> {
public:
  using Super = GVarProp_<bool> ;
  using IsBool = bool;

  GBool();
  GBool(bool value);
  GBool(GString const& value);  ///< true if value = "true" or "yes", false otherwise.
  GBool(GBool const&);
  GBool* clone(bool copy_value) const override;

  GBool* toBool() override {return this;}
  bool   boolValue()  const override {return value_;}
  float  floatValue() const override {return float(value_);}

  GBool& operator=(bool val) {set(val,false); return *this;}
  GBool& operator=(GBool const& val) {set(val); return *this;}
  GBool& operator=(GString const& val) {set(val); return *this;}

  using Super::set;
  void set(GProp const&) override;

  /// Future value in an Active Expression (see GExpr).
  auto operator~() {return GPropRef_<GBool>(*this);}

  /// Returns a trigger testing the state value that can be used in a Conditional Expression.
  /// - +state: equivalent to (~state == true) : fires callbacks and activates conditional props
  ///  when the state is entered; deactivates conditional props when the state is exited;
  /// - -state: equivalent to  (~state == false)
  /// @code
  ///    auto& status = Bool();
  ///    mygadget
  ///    << +status / [this]{doAction()
  ///    << +status / "OK" / GColor.green
  ///    << -status / "Wrong" / GColor.red;
  /// @endcode
  /// @{
  GBoolExpr& operator+();
  GBoolExpr& operator-();
  /// @}
  ///
  /// Binds this model to an Active Expression (see GExpr).
  GBool& operator<<=(GBoolExpr&);

  using GType = GPropType_<GBool,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Return a GBool created in the heap.
/// @{
GBool& Bool();
inline GBool& Bool(bool value) {return *new GBool(value);}
GBool& Bool(GString const& value);
/// @}

}
#endif
