//
// Active Expressions (auto-reevaluated)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <regex>
#include <gevent.hpp>
#include <gnotifier.hpp>
#include <gbind.hpp>
#include <gexpr.hpp>
#include <gbool.hpp>
#include <gint.hpp>
#include <gfloat.hpp>
#include <gtextitem.hpp>
#include <gbox.hpp>
#include <gnumber.hpp>
#include <gstatechart.hpp>
#include <gcontrol.hpp>
using namespace std;
namespace guit {

GPropType& GExpr::Type() {
  static auto& t(*new GPropType("Expr", GPropType::Expr));
  return t;
}

GPropType& GExprWrapper::Type() {
  static auto& t(*new GPropType("ExprWrapper", GPropType::ExprWrapper));
  return t;
}

class GBoolFun : public GProp {
public:
  GBoolFun(std::function<bool()> const& fun) : fun_(fun) {}

  bool boolValue() const override {return (fun_)();}
  float floatValue() const override {return boolValue();}
  GString stringValue() const override {return to_string(boolValue());}

protected:
  std::function<bool()> fun_{};
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GExprBool::GExprBool(bool val) : prop(*new GBool(val)) {}
GExprBool::GExprBool(GBool& val) : prop(val) {}
GExprBool::GExprBool(GPropRef_<GBool> const& ref) : prop(ref.prop_) {}
GExprBool::GExprBool(GBoolExpr& expr) : prop(expr) {}
GExprBool::GExprBool(std::function<bool()> const& fun) : prop(*new GBoolFun(fun)) {}

GExprText::GExprText(const char* val) : prop(*new GText(val)) {}
GExprText::GExprText(GString const& s) : prop(*new GText(s)) {}
GExprText::GExprText(GPropRef const& ref) : prop(ref.prop_) {}
GExprText::GExprText(GTextExpr& expr) : prop(expr) {}

GExprNumber::GExprNumber(int val) : prop(*new GInt(val)) {}
GExprNumber::GExprNumber(float val) : prop(*new GFloat(val)) {}

static float checkDouble(double from) {float to; gconvert(from, to); return to;}

GExprNumber::GExprNumber(double val) : prop(*new GFloat(checkDouble(val))) {}
GExprNumber::GExprNumber(GNumPropRef const& ref) : prop(ref.prop_) {}
GExprNumber::GExprNumber(GNumExpr& expr) : prop(expr) {}
GExprNumber::GExprNumber(GBoolExpr& expr) : prop(expr) {}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GExpr::GExpr(GProp const* left, Oper oper, GProp const* right) :
left_(const_cast<GProp*>(left)),
right_(const_cast<GProp*>(right)),
oper_(oper) {}

GExpr::GExpr(const GExpr& from)
: left_(from.left_), right_(from.right_), oper_(from.oper_) {}

void GExpr::bindImpl(GExprWrapper& wrap) {
  if (!left_) ;
  else if (auto* e = left_->toExpr())
    e->bindImpl(wrap);
  else if (auto* m = left_->toMutableProp())
    m->add(*wrap.notifier_);
  else if (auto* l = left_->toPropWrapper()) {
    if (auto* ml = l->prop()->toMutableProp()) ml->add(*wrap.notifier_);
  };
  
  if (!right_) ;
  else if (auto* e = right_->toExpr())
    e->bindImpl(wrap);
  else if (auto* m = right_->toMutableProp())
    m->add(*wrap.notifier_);
  else if (auto* r = right_->toPropWrapper()) {
    if (auto* mr = r->prop()->toMutableProp()) mr->add(*wrap.notifier_);
  };
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GBoolExpr::GBoolExpr(GProp const* left, Oper oper, GProp const* right, Type type)
: GExpr(left, oper, right), type_(type) {}

GBoolExpr::GBoolExpr(const GBoolExpr& from)
: GExpr(from), type_(from.type_) {}

void GBoolExpr::linkWithTrigger(GExprTrigger& trigger) {
  linkImpl(trigger, *this, true/* add notifier*/);
}

GString GBoolExpr::stringValue() const {
  GString txt;
  gconvert(boolValue(), txt);
  return txt;
}

bool GBoolExpr::boolValue() const {
  if (type_ == TBool) {
    bool a{}, b{};
    bool aset{}, bset{};
    
    if (left_) {
      a = left_->boolValue();
      aset = true;
    }
    if (!aset) goto WrongArgument;
    
    switch (oper_) {
      case Identity:
        return a;
      case Not:
        return !a;
      case Or:
        if (a == true) return true;
        break;
      case And:
        if (a == false) return false;
        break;
      default:
        goto WrongOperator;
    }
    
    if (right_) {
      b = right_->boolValue();
      bset = true;
    }
    if (!bset) goto WrongArgument;
    return b;
  }
  
  else if (type_ == TNum) {
    float a{}, b{};
    bool aset{}, bset{};
    
    if (left_) {a = left_->floatValue(); aset = true;}
    if (right_) {b = right_->floatValue(); bset = true;}
    
    if (!aset || !bset) goto WrongArgument;
    switch (oper_) {
      case Equal:
        return a == b;
      case NotEqual:
        return a != b;
      case Greater:
        return a > b;
      case GreaterEqual:
        return a >= b;
      case Less:
        return a < b;
      case LessEqual:
        return a <= b;
      default:
        goto WrongOperator;
    };
  }
  
  else if (type_ == TText) {
    GString a, b;
    bool aset{}, bset{};
    
    if (left_) {a = left_->stringValue(); aset = true;}
    if (right_) {b = right_->stringValue(); bset = true;}
    
    if (!aset || !bset) goto WrongArgument;
    switch (oper_) {
      case Equal:
        return a == b;
      case NotEqual:
        return a != b;
      case Greater:
        return a > b;
      case GreaterEqual:
        return a >= b;
      case Less:
        return a < b;
      case LessEqual:
        return a <= b;
      case StrStart:
        return a.compare(0, std::string::npos, b, 0, std::string::npos);
      case StrContain:
        return a.find(b);
      case StrRegex: {
        try {
          std::regex pattern{b};
          return std::regex_match(a, pattern);
        }
        catch (...) {
          return false;
        }
      }
      default:
        goto WrongOperator;
    };
  }
WrongOperator:
  error("compute","Invalid boolean expression: Wrong operator");
  return false;
  
WrongArgument:
  error("compute","Invalid boolean expression: Wrong arguments");
  return false;
}


GBoolExpr& operator!(GExprBool const& a) {
  return *new GBoolExpr(&a.prop, GExpr::Not, nullptr, GBoolExpr::TBool);
}

GBoolExpr& operator&&(GExprBool const& a, GExprBool const& b) {
  return *new GBoolExpr(&a.prop, GExpr::And, &b.prop, GBoolExpr::TBool);
}

GBoolExpr& operator&&(GExprBool const& a, std::function<bool()> const& fun) {
  return *new GBoolExpr(&a.prop, GExpr::And, new GBoolFun(fun), GBoolExpr::TBool);
}

GBoolExpr& operator||(GExprBool const& a, GExprBool const& b) {
  return *new GBoolExpr(&a.prop, GExpr::Or, &b.prop, GBoolExpr::TBool);
}

GBoolExpr& operator||(GExprBool const& a, std::function<bool()> const& fun) {
  return *new GBoolExpr(&a.prop, GExpr::Or, new GBoolFun(fun), GBoolExpr::TBool);
}

GBoolExpr& operator==(GExprBool const& a, GExprBool const& b) {
  return *new GBoolExpr(&a.prop, GExpr::Equal, &b.prop, GBoolExpr::TNum);
}

GBoolExpr& operator!=(GExprBool const& a, GExprBool const& b) {
  return *new GBoolExpr(&a.prop, GExpr::NotEqual, &b.prop, GBoolExpr::TNum);
}


GBoolExpr& operator==(GExprNumber const& a, GExprNumber const& b)
{return *new GBoolExpr(&a.prop, GExpr::Equal, &b.prop, GBoolExpr::TNum);}

GBoolExpr& operator!=(GExprNumber const& a, GExprNumber const& b)
{return *new GBoolExpr(&a.prop, GExpr::NotEqual, &b.prop, GBoolExpr::TNum);}

GBoolExpr& operator>(GExprNumber const& a, GExprNumber const& b)
{return *new GBoolExpr(&a.prop, GExpr::Greater, &b.prop, GBoolExpr::TNum);}

GBoolExpr& operator>=(GExprNumber const& a, GExprNumber const& b)
{return *new GBoolExpr(&a.prop, GExpr::GreaterEqual, &b.prop, GBoolExpr::TNum);}

GBoolExpr& operator<(GExprNumber const& a, GExprNumber const& b)
{return *new GBoolExpr(&a.prop, GExpr::Less, &b.prop, GBoolExpr::TNum);}

GBoolExpr& operator<=(GExprNumber const& a, GExprNumber const& b)
{return *new GBoolExpr(&a.prop, GExpr::LessEqual, &b.prop, GBoolExpr::TNum);}

// - - - -

GBoolExpr& operator==(GExprText const& a, const GExprText& b) {
  return *new GBoolExpr(&a.prop, GExpr::Equal, &b.prop, GBoolExpr::TText);
}

// true if first argument starts with second argument.
GBoolExpr& operator^=(GExprText const& a, GExprText const& b) {
  return *new GBoolExpr(&a.prop, GExpr::StrStart, &b.prop, GBoolExpr::TText);
}

// true if first argument contains second argument.
GBoolExpr& operator*=(GExprText const& a, GExprText const& b) {
  return *new GBoolExpr(&a.prop, GExpr::StrContain, &b.prop, GBoolExpr::TText);
}

// true if first argument matches the regex specified in the second argument.
GBoolExpr& operator%=(GExprText const& s, GExprText const& regex) {
  return *new GBoolExpr(&s.prop, GExpr::StrRegex, &regex.prop, GBoolExpr::TText);
}

GBoolExpr& operator!=(GExprText const& a, GExprText const& b) {
  return *new GBoolExpr(&a.prop, GExpr::NotEqual, &b.prop, GBoolExpr::TText);
}

GBoolExpr& operator>(GExprText const& a, GExprText const& b) {
  return *new GBoolExpr(&a.prop, GExpr::Greater, &b.prop, GBoolExpr::TText);
}

GBoolExpr& operator>=(GExprText const& a, GExprText const& b) {
  return *new GBoolExpr(&a.prop, GExpr::GreaterEqual, &b.prop, GBoolExpr::TText);
}

GBoolExpr& operator<(GExprText const& a, GExprText const& b) {
  return *new GBoolExpr(&a.prop, GExpr::Less, &b.prop, GBoolExpr::TText);
}

GBoolExpr& operator<=(GExprText const& a, GExprText const& b) {
  return *new GBoolExpr(&a.prop, GExpr::LessEqual, &b.prop, GBoolExpr::TText);
}

GTextExpr& operator+(GExprText const& a, GExprText const& b) {
  return *new GTextExpr(&a.prop, GExpr::Add, &b.prop);
}

GTextExpr& operator+(GExprText const& a, GExprNumber const& b) {
  return *new GTextExpr(&a.prop, GExpr::Add, &b.prop);
}

GTextExpr& operator+(GExprNumber const& a, GExprText const& b) {
  return *new GTextExpr(&a.prop, GExpr::Add, &b.prop);
}

// - - - -

static GInt& zero = *new GInt{0};
static GInt& one = * new GInt{1};

GBoolExpr& GBool::operator+() {
  return *new GBoolExpr(this, GExpr::Equal, &one, GBoolExpr::TNum);
}

GBoolExpr& GBool::operator-() {
  return *new GBoolExpr(this, GExpr::Equal, &zero, GBoolExpr::TNum);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool GTextExpr::boolValue() const {
  auto s = stringValue();
  bool val = 0;
  gconvert(s, val);
  return val;
}

float GTextExpr::floatValue() const {
  return stof(stringValue());
}

GString GTextExpr::stringValue() const {
  GString a, b;
  bool aset{}, bset{};
  
  if (left_) {a = left_->stringValue(); aset = true;}
  else error("toString","Invalid left operand");
  
  if (right_) {b = right_->stringValue(); bset = true;}
  else error("toString","Invalid right operand");
  
  if (!aset || !bset) {
    error("toString","Invalid textual expression: Wrong argument");
    return "";
  }
  
  switch (oper_) {
    case Add:
      return a + b;
      //case Assign:
      //  if (l) l->setFromProp(Text(b));
      //  return b;
    default:
      error("toString","Invalid textual expression: Wrong operator");
      return "";
  }
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GString GNumExpr::stringValue() const {
  GString txt;
  gconvert(floatValue(), txt);
  return txt;
}

float GNumExpr::floatValue() const {
  float a = 0.f, b = 0.f;
  bool aset{}, bset{};
  
  if (left_) {a = left_->floatValue(); aset = true;}
  else error("computeFloat","Invalid left operand");
  
  if (right_) {b = right_->floatValue(); bset = true;}
  else error("computeFloat","Invalid right operand");
  
  if (!aset || !bset) {
    error("computeFloat","Invalid numeric expression: Wrong argument");
    return 0.f;
  }
  
  switch (oper_) {
      //case Assign:
      //  if (l) l->setFromProp(GFloat(b));
      //  return b;
    case Add:
      return a + b;   // actual types not taken into account, always float
    case Substract:
      return a - b;
    case Multiply:
      return a * b;
    case Divide:
      if (b == 0.f) {
        gerror("GExpr","division by 0");
        return std::numeric_limits<float>::max();
      }
      return a / b;  // no integral division!!!
    default:
      error("compute","Invalid numeric expression: Wrong operator");
      return 0.f;
  };
}

GNumExpr& operator+(GExprNumber const& a, GExprNumber const& b) {
  return *new GNumExpr(&a.prop, GExpr::Add, &b.prop);
}

GNumExpr& operator-(GExprNumber const& a, GExprNumber const& b) {
  return *new GNumExpr(&a.prop, GExpr::Substract, &b.prop);
}

GNumExpr& operator*(GExprNumber const& a, GExprNumber const& b) {
  return *new GNumExpr(&a.prop, GExpr::Multiply, &b.prop);
}

GNumExpr& operator/(GExprNumber const& a, GExprNumber const& b) {
  return *new GNumExpr(&a.prop, GExpr::Divide, &b.prop);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GExprTrigger::~GExprTrigger() {
  deleting_ = true;
  for (auto& it : senders_)  {if (it) it->removeNotifiers(this);}
}

GExprTrigger::GExprTrigger(GBoolExpr& e) {
  boolexpr_ = e;
  boolexpr_->linkWithTrigger(*this);
}

bool GExprTrigger::operator==(GTrigger const& other) const {
  //cerr << "debug: GExprTrigger::operator== " << endl;
  return false;
  //isEventTrigger() && other.isEventTrigger()
  //&& mask_ == other.mask_ && filter_ == other.filter_;
}

void GExprTrigger::setNotifier(GNotifier* n) {
  notifier_ = n;
}

void GExprTrigger::fire(GObject& sender, bool active) {
  if (!recv_  || invalid_ || !firedprop_) return;
  GEvent e(GTrigger(GEventType::Change, active), GCore::elapsedTime());
  //GCore::setCurrentEvent(e);
  e.gadget_ = recv_;
  e.aux_ = &sender;
  firedprop_->activate(e, active);
  //GCore::setCurrentEvent(nullptr);
}

void GExprTrigger::addSender(GObject& sender) {
  auto it = std::find(senders_.begin(), senders_.end(), &sender);
  if (it == senders_.end()) {
    senders_.insert_after(senders_.before_begin(), &sender);
  }
}

void GExprTrigger::senderRemoved(GObject& obj) {
  invalid_ = true;
  auto prev_it = senders_.before_begin();  // remove all occurences of obj
  auto it = senders_.begin();
  while (it != senders_.end()) {
    if (*it == &obj) {
      if (deleting_) *it = nullptr;  // utile?
      else {
        ++it;
        senders_.erase_after(prev_it);
        continue;
      }
    }
    prev_it = it++;
  }
  
  if (deleting_) return;
  // remove all senders (if any other)
  deleting_ = true;
  for (auto& it2 : senders_) {if (it2) it2->removeNotifiers(this);}
  senders_.clear();
}

void GExprTrigger::receiverAdded(Gadget& g, GCond& p) {
  recv_ = &g;
  firedprop_ = &p;
  // active la prop selon la valeur initiale de la prop
  if (p.isCondProp() && !invalid_ && notifier_) {
    GEvent e(GTrigger(GEventType::Change, -1), GCore::elapsedTime());
    e.aux_ = &p;
    notifier_->fire(e);
  }
}

void GExprTrigger::receiverRemoved() {
  recv_ = nullptr;
}

void GExprTrigger::receiverDeleted() {
  //                          traiter guard_ !!!!!!
  recv_ = nullptr;
  deleting_ = true;
  for (auto& it : senders_) if (it) it->removeNotifiers(this);
  senders_.clear();
  delete this;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GExprNotifier::fire(GEvent&)  {
  if (!wrap.model_) return;
  switch (wrap.type_) {
    case GExprWrapper::TBool:
    case GExprWrapper::TNum:
      wrap.model_->set(*wrap.expr_);
      break;
    case GExprWrapper::TString:
      wrap.model_->set(wrap.expr_->stringValue());
      break;
    case GExprWrapper::TText:
      wrap.model_->set(wrap.expr_->stringValue());
      break;
    default:
      // nop
      break;
  }
}

GExprWrapper::GExprWrapper(GExpr& expr) :
expr_(expr),
notifier_(new GExprNotifier(*this)),
model_{} {
  //cerr << "GExprWrapper::GExprWrapper "<<this <<" expr" <<&expr<< endl;
  expr_->bindImpl(*this);
}

float GExprWrapper::floatValue() const  {
  return expr_->floatValue();
}

GString GExprWrapper::stringValue() const  {
  return expr_->stringValue();
}

void GExprWrapper::removeNotifiers(void* obj) {
  //cerr << "GExprWrapper::removeNotifiers "<<this <<" model "<<model_ << endl;
  if (obj == model_) model_ = nullptr;
}

void GExprWrapper::onRemove(Gadget*) {
  //cerr << "GExprWrapper::onRemove " << this <<endl;
  if (model_) model_->removeNotifiers(this);
  model_ = nullptr;
}

bool GExprWrapper::bindWith(GVarProp& prop) {
  // cerr << "GExprWrapper::bindWith: expr: "<<this
  // <<" prop: " << &prop<< " mutable: " << prop.isMutable() << endl;
  type_ = TNone;
  
  if (expr_->isTextExpr()) {
    type_ = TString;
  }
  else if (expr_->isBoolExpr()) {
    if (prop.toText()) type_ = TString;
    else if (prop.toBool()) type_ = TBool;
    else {
      error("bindWith","This prop not compatible with a boolean expression");
      return false;
    }
  }
  else if (expr_->isNumExpr()) {
    if (prop.toText()) type_ = TText;   // !! Text vs String !!
    else if (prop.toInt() || prop.toFloat() || prop.toBool()) type_ = TNum;
    else {
      error("bindWith","This prop not compatible with a numeric expression");
      return false;
    }
  }
  
  if (type_ != TNone) {
    model_ = &prop;
    GEvent e(GTrigger::none, GCore::elapsedTime());
    e.aux_ = this;
    notifier_->fire(e); // init
    return true;
  }
  else {
    error("bindWith","Type of model and expression are not compatible");
    return false;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GExpr::linkImpl(GExprTrigger& trigger, GBoolExpr& fullexpr, bool add_notifier) {
  
  // add createDeleteNotifier(s) ??
  
  if (left_) {
    if (auto* e = left_->toExpr()) {
      e->linkImpl(trigger, fullexpr, add_notifier);
    }
    else if (auto* m = left_->toMutableProp()) {
      trigger.addSender(*m);
      if (add_notifier)
        m->add(*new GTriggerExprNotifier(*left_, trigger, fullexpr));
    }
  }

  if (right_) {
    if (auto* e = right_->toExpr()) {
      e->linkImpl(trigger, fullexpr, add_notifier);
    }
    else if (auto* m = right_->toMutableProp()) {
      trigger.addSender(*m);
      if (add_notifier)
        m->add(*new GTriggerExprNotifier(*right_, trigger, fullexpr));
    }
  }
}

}
