//
// Base classes of Properties (Props in short)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gfile.hpp>
#include <gpropspec.hpp>
#include <gcontrol.hpp>
#include <gstyle.hpp>
#include <gadgetstyle.hpp>
using namespace std;
namespace guit {

using TypeMap = std::unordered_map<GString,GPropType*>;

GPropType::GPropType(GString const& name, int role)
: name_(name), role_(role) {
  addType();
}

GPropType::GPropType(GString const& name, GPropType& parent)
: name_(name), role_(parent.role()), parent_(&parent) {
  addType();
}

bool GPropType::operator==(GPropType const& other) const {
  return this == &other;
}

bool GPropType::operator!=(GPropType const& other) const {
  return this != &other;
}

static TypeMap& propTypes() {
  static TypeMap* defs = new TypeMap;
  return *defs;
}

 GPropType* GPropType::findType(GString const& type) {
   auto& defs = propTypes();
   auto it = defs.find( strlower(type) );
   return (it == defs.end()) ? nullptr : it->second;
 }

GProp* GPropType::createInstance(GString const& type, GString const& arg,
                                  bool css_style) {
  size_t pos = 0;
  if (!css_style
      || (pos = type.find('-')) == GString::npos
      || pos+1 >= type.size()
      ) {
    if (auto* cl = findType(type)) {
      // must have same case if not in CSS
      if (css_style || cl->name() == type) return &cl->newInstance(arg);
    }
  }
  else {
    auto type2 = type.substr(0, pos);
    auto arg2 = type.substr(pos+1) + "=" + arg;
    if (auto* cl = findType(type2)) return &cl->newInstance(arg2);
  }
  return nullptr;
}

void GPropType::addType() {
  static int roleCount_ = GPropType::Role::Auto;
  auto& defs = propTypes();
  auto it = defs.find(strlower(name_));

  if (it != defs.end()) {
    gerror("GPropType","This Type ("+name_+") already exists (ignored)");
    return;
  }
  if (role_ == Role::Auto) role_ = ++roleCount_;
  //defs.insert_or_assign(strlower(name_), this);  c++17
  defs[strlower(name_)] = this;
}

GProp& GPropType::newInstance()  {
  throw GFatalError("GPropType::newInstance",
                    "This Type (" + name() + ") can't be instantiated");
}

GProp& GPropType::newInstance(GString const& arg)  {
  throw GFatalError("GPropType::newInstance",
                    "This Type (" + name() + ") can't be instantiated");
}

void GPropType::operator delete(void* ptr) {
  gerror("GPropType::delete",
         "The delete operator can't be used to destroy a Type (ignored)");
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// needed beacuse this class is abstract and has no actual Type
// (a virtual pure fct exception may be raised otherwise if there is an error)
GPropType& GProp::Type() {
  static GPropType& t = *new GPropType("Prop", GPropType::Abstract);
  return t;
}

GString GProp::typeName() const {return type().name();}

int GProp::role() const {return type().role();}

bool GProp::isEquivalent(GProp const& other) const {
  // !NOTE: props with a negative role are never equivalent !!
  return role() > GPropType::Abstract && role() == other.role();
}

bool GProp::onAdd(Gadget* g) {
  if (g) addModes(g->modes_);
  return true;
}

void GProp::onRemove(Gadget* g) {
  if (g) removeModes(g->modes_);
}

std::ostream& operator<<(std::ostream& out, GProp const& prop) {
  prop.write(out); return out;
}

bool GProp::write(GString const& filename) const {
  GOFile out(filename);
  if (!out) return false;
  write(out);
  return true;
}

void GProp::write(std::ostream& out) const {
  out << stringValue();
}

std::istream& operator>>(std::istream& in, GVarProp& prop) {
  prop.read(in);
  return in;
}


GFieldSpecValues::GFieldSpecValues(std::initializer_list<GFieldSpecValue> list)
: list_(list) {
  for (auto& it : list_) {
    if (!it.name_.empty()) {
      it.nameid_ = strid(strlower(it.name_));
      if (it.name_[0] != '%') it.islabel_ = true;
    }
  }
}

GFieldSpecValue const* GFieldSpecValues::find1(GString const& str) const {
  if (str.empty()) return nullptr;
  uint64_t id = strid(strlower(str));
  for (auto& it : list_) {
    if (it.nameid_ == id) return &it;
  }
  return nullptr;
}

GFieldSpecValue const* GFieldSpecValues::find2(GString const& str, float& val) const {
  if (str.empty()) return nullptr;

  if (!::isdigit(str[0]) && str[0] != '-' && str[0] != '+' && str[0] != '.') {
    auto it = find1(str);
    if (!it) return nullptr;
    else {
      val = it->value_;
      return it;
    }
  }
  else {
    char* pend{};
    val = std::strtof(str.data(), &pend);  // nb: locale set in GPropSpec::toValue()
    /*
     cerr << "find2: '" <<str <<"' val: "<<val
     << " / pend: " << *pend <<" " << int(*pend)
     << " / error: " << (str.data() == pend) << endl;
     */
    if (pend == str.data())  // error: no conversion could be performed
      return nullptr;
    else if (*pend == 0)      // just a number: no suffix
      return find1("%f");
    else
      return find1(GString("%") + pend);   // check if suffix is allowed
  }
}


bool GFieldSpecValues::find1(int type, GString& to) const {
  for (auto& it : list_) {
    if (!it.isalias_ && it.type_ == type) {
      to = it.name_;
      return true;
    }
  }
  to.clear();
  return false;
}

bool GFieldSpecValues::find2(int type, float val, GString& to) const {
  GFieldSpecValue const* found{};

  for (auto& it : list_) {
    if (!it.isalias_ && it.type_ == type) {
      found = &it;
      break;
    }
  }

  if (!found) {
    to.clear();
    return false;
  }

  if (found->name_[0] != '%') to = found->name_;
  else {
    gconvert(val, to);
    if (found->name_ != "%f" && found->name_.size() >= 2) {
      to += found->name_.substr(1);
    }
  }
  return true;
}


GPropSpec::GPropSpec(std::initializer_list<GFieldSpec*> fields)
: fields_(fields) {}


int GPropSpec::toValue(GString const& str, Value* prop) {
  GString curloc = setlocale(LC_ALL,NULL);
  setlocale(LC_NUMERIC, "C");  // needed by gconvert()
  //bool stat = true;

  // strlower(str) !!

  size_t pos{};
  GString tok, tok2;
  int del{};
  size_t count = 0;

  while ((del = strtoken(str, tok, pos, " \t<\n/")) != -1) {

    if (count >= fields_.size()) {  // too many fields
      //stat = false;
      count = -1;
      break;
    }

    if (!fields_[count++]->toValue(tok, prop)) {
      //stat = false;   // wrong field
      count = -1;
      break;
    }

    if (del != '/' && count < fields_.size() && fields_[count]->isoptional_) {
      fields_[count++]->toValue(tok, prop);
    }
  }

  setlocale(LC_NUMERIC, curloc.data());
  return int(count);
}


GString GPropSpec::toString(const Value* prop) {
  GString curloc = setlocale(LC_ALL,NULL);
  setlocale(LC_NUMERIC, "C");  // needed by gconvert()

  std::ostringstream ss;
  //ss.imbue(std::locale("C"));
  ss.precision(GPrefs::precision());

  bool first = true;

  for (size_t k = 0; k < fields_.size(); ++k) {

    if (fields_[k]->isoptional_) ss << '/';
    else if (!first) ss << ' ';

    GString str;
    fields_[k]->toString(prop, str);
    ss << str;

    first = false;
  }

  setlocale(LC_NUMERIC, curloc.data());
  return ss.str();
}

}
