//
// Base class of Properties (Props in short)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Prop_hpp
#define Guit_Prop_hpp
#include <gobject.hpp>
#include <gproptype.hpp>
namespace guit {

/// Base class of Properties.
/// Properties specify the appearance or behavior of the gadgets they are
/// added to. They can be shared by several gadgets and are added to them
/// by using the << operator. See Gadget for details and examples.
///
/// The class has two main subclasses: GCond for conditional expressions
/// (Callbacks and Conditional Props) and GVarProp for all other props.
///
/// @see Gadget, Callback, GCond.
class GProp : public GObject {
public:
  using IsProp = bool;
  using IsGadgetChild = bool;

  /// Specificity of Properties.
  using Specif = int32_t;

  /// Clones the prop (when applicable).
  /// Mutable props (i.e. deriving from GVarProp) can be cloned, otherwise this method returns null.
  /// The value is copied if _copy_value_ is used, the default value is used otherwise.
  virtual GProp* clone(bool copy_value) const {return nullptr;}

  /// Returns prop's meta-class.
  /// Both functions return the prop's type, but Type() is a static method and type() a virtual method.
  /// @{
  using   GType = GPropType;
  static  GType& Type();
  virtual GType& type() const {return Type();}
  /// @}

  /// Returns the name and role of the prop's type.
  /// @{
  GString typeName() const override;
  int role() const;
  /// @}

  /// Class conversions.
  /// @{
  GProp* toProp() override {return this;}
  GProp const* toProp() const override {return this;}
  /// @}

  /// Returns true if the prop is mutable.
  /// Returns true only if the prop derives from GVarProp and setImmutable() was not called.
  /// Props with another type than GVarProp are always immutable.
  virtual bool isMutable() const {return false;}

  /// Makes the prop immutable.
  /// Affects only GVarProp props (other props are always immutable).
  /// A prop cannot be made mutable again after this method has been called.
  virtual void setImmutable() {}

  virtual bool isString() const {return false;}
  virtual bool isNumber() const {return false;}
  virtual bool isCondProp() const {return false;}

  /// Returns the value of the prop as a string.
  /// Returns "" if the prop has no value of its value cannot be converted to a string.
  virtual GString stringValue() const {return "";}

  /// Returns the value of the prop as a float.
  /// Returns 0.f if the prop has no value of its value cannot be converted to a float.
  virtual float floatValue() const {return 0.f;}

  /// Returns the value as a bool.
  /// Returns false if the prop has no value of its value cannot be converted to a bool.
  virtual bool boolValue() const {return false;}

  /// Writes the value of the prop on a file or a stream.
  /// @{
  bool write(GString const& filename) const;
  virtual void write(std::ostream& out) const;
  friend std::ostream& operator<<(std::ostream& out, GProp const&);
  /// @}

  /// Returns true if these props are equivalent.
  /// Adding a prop to a gadget replaces the equivalent prop it contains (if it contains any).
  /// In all other cases, the prop is added to the gadget proplist.
  /// NOTE: props with a negative role are never equivalent!
  virtual bool isEquivalent(GProp const& other) const;

  /// Called when the property is added to an object.
  /// note: must call addModes().
  virtual bool onAdd(Gadget*);

  /// Called when the property is removed from an object.
  /// note: must call removeModes().
  virtual void onRemove(Gadget*);

  virtual void addModes(struct GadgetModes&) {}
  virtual void removeModes(struct GadgetModes&) {}
  virtual void apply(Gadget*, GRender&, Specif) {}
  virtual void fire(GEvent&) {}
  virtual void updateGadget(Gadget&) {}

  /// Type conversions.
  /// @{
  virtual class GVarProp*     toVarProp() {return nullptr;}
  virtual class GVarProp*     toMutableProp() {return nullptr;}
  virtual class GText*        toText() {return nullptr;}
  virtual class GBool*        toBool() {return nullptr;}
  virtual class GNumber*      toNumber()  {return nullptr;}
  virtual class GInt*         toInt()  {return nullptr;}
  virtual class GFloat*       toFloat(){return nullptr;}
  virtual class GPos*         toPos()  {return nullptr;}
  virtual class GPos const*   toPos()  const {return nullptr;}
  virtual class GSize*        toSize() {return nullptr;}
  virtual class GSize const*  toSize() const {return nullptr;}
  virtual class GChoice*      toChoice() {return nullptr;}
  virtual class GAnim*        toAnim() {return nullptr;}
  virtual class GCond*        toCond() {return nullptr;}
  virtual class GExpr*        toExpr() {return nullptr;}
  virtual class GBoolExpr*    toBoolExpr() {return nullptr;}
  virtual class GExprWrapper* toExprWrapper() {return nullptr;}
  virtual class GPropWrapper* toPropWrapper() {return nullptr;}
  virtual class GNotifier*    toNotifier() {return nullptr;}
  /// @}

protected:
  GProp() = default;
  GProp(GProp const&) = delete;
  GProp(GProp const&&) = delete;
  GProp& operator=(GProp const&) = delete;
  GProp& operator=(GProp const&&) = delete;
};

}
#endif 
