//
// Text, RegText and Tip Props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <regex>
#include <gtext.hpp>
#include <gexpr.hpp>
#include <gint.hpp>
#include <gfloat.hpp>
#include <gbox.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gtextrender.hpp>
#include <gstring.hpp>
#include <gadgetstyle.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP(Text,GPropType::Text)
GIMPLEMENT_PROP(RegText,GPropType::Text)
GIMPLEMENT_PROP(Tip,GPropType::Tip)
GIMPLEMENT_PROP(TextData,GPropType::Auto)

GTextData* GTextData::clone(bool copy_value) const {
  return copy_value ? new GTextData(value_) : new GTextData;
}

GText::GText() {}

GText::GText(GString const& value) : Super(value) {}

GText::GText(GText const& from) : Super(from.value_) {}

GText::GText(int value) : GText() {
  pmodes_.isSet = 1;
  gconvert(value, value_);
}

GText::GText(float value) : GText() {
  GString str;
  gconvert(value, str);
  set(str);
}

GText* GText::clone(bool copy_value) const {
  return copy_value ? new GText(*this) : new GText;
}

bool GText::onAdd(Gadget* g) {
  if (g && !empty() && value_[0] == '{') {
    GString rest;
    GadgetStyle::initStyle(*g, value_, rest);
    //if (!title.empty()) value_->set(title);
  }
  return true;
}

void GText::onRemove(Gadget*) {
}

float GText::floatValue() const {
  return strtof(value().data(), nullptr);
}

GText& GText::operator<<=(GExpr& expr) {
  bindImpl(expr);
  set(expr.stringValue(), false);
  return *this;
}

GText& GText::operator=(GInt const& num) {
  set(num.stringValue(), false);
  return *this;
}

GText& GText::operator=(GFloat const& num) {
  append(num.stringValue());
  return *this;
}

GText& GText::operator+=(GInt const& num) {
  append(num.stringValue());
  return *this;
}

GText& GText::operator+=(GFloat const& num) {
  set(num.stringValue(), false);
  return *this;
}

GString operator+(GString const& str, GText const& text) {
  return str + text();
}

GString operator+(GText const& text, GString const& str) {
  return text() + str;
}

GString operator+(GText const& text, GInt& num) {
  return text() + num.stringValue();
}

GString operator+(GText const& text, GFloat& num) {
  return text() + num.stringValue();
}

void GText::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Text]) return;
  r.specif[GPropType::Text] = specif;

  r.text_ = &value_;
}

void GText::updateGadget(Gadget& g) {
  //cerr << "GText::updateGadget: " << value_ << endl;
  g.updateLayout();
}

void GText::set(GProp const& prop) {
  set(prop.stringValue(), false);
}

bool GText::empty() const {return value_.empty();}

size_t GText::size() const {return value_.size();}

int GText::compare(GString const& str) const noexcept {return value_.compare(str);}

size_t GText::find(GString const& str, size_t pos) const noexcept {
  return value_.find(str, pos);
}

size_t GText::rfind(GString const& str, size_t pos) const noexcept {
  return value_.rfind(str, pos);
}

GString GText::substr(size_t pos, size_t len) const {
  return value_.substr(pos, len);
}

char GText::operator[](size_t pos) const {      // !!! att unicode !!!!
  if (pos < 0 || pos >= size()) return 0;
  else return value_[pos];
}

bool GText::setByteAt(int pos, char ch) {   // !!! att unicode !!!!
  if (pos < 0 || size_t(pos) >= size()) return false;
  GString newval(value_);
  newval[pos] = ch;
  return edit(newval);
}

bool GText::append(GString const& str) {
  GString newval(value_);
  newval.append(str);
  return edit(newval);
}

bool GText::erase(size_t pos, size_t len) {
  if (pos >= value_.size()) return false;
  GString newval(value_);
  newval.erase(pos, len);
  return edit(newval);
}

bool GText::insert(size_t pos, GString const& arg) {
  if (pos > value_.size()) return false;
  GString newval(value_);
  newval.insert(pos, arg);
  return edit(newval);
}

bool GText::insert(size_t pos, GString const& arg, size_t subpos, size_t sublen) {
  if (pos > value_.size()) return false;
  GString newval(value_);
  newval.insert(pos, arg, subpos, sublen);
  return edit(newval);
}

void GText::read(std::istream& in) {
  if (!isMutable()) {immutableError("read()"); return;}

  value_.assign(std::istreambuf_iterator<char>(in),
                std::istreambuf_iterator<char>());
  bool istext = true;

  if (value_.size() > 5 && value_[0]=='%' && value_[1]=='P'
      && value_[2]=='D' && value_[3]=='F')
    istext = false;
  else {
    for (size_t k=0; k < value_.size() && k < 500; k++) {
      if (std::iscntrl(value_[k]) && !std::isspace(value_[k])) {
        istext = false;
        break;
      }
    }
  }

  if (istext) set(value_,true);
  else {
    value_.clear();
    set(value_,true);
  }
}

bool GText::edit(GString const& newval) {
  if (!isValid(newval)) return false;
  
  if (!pmodes_.hasMaster) value_ = newval;
  else {
    GString copy{value_};
    value_ = newval;
    if (!testMasterValue()) {
      value_ = copy;
      return false;
    }
  }
  pmodes_.changing = 1;
  fireDefaultCallbacks();
  pmodes_.changing = 0;
  return true;
}

bool GText::doChange(GString const& s) {
  if (!isValid(s) || (s == value_ && pmodes_.isSet)) return false;
  pmodes_.isSet = 1;
  value_ = s;
  return true;
}

GDim GText::textSize(GFont& f, bool richtext, bool multilines,
                     GTextAttributes const* attrs) const {
  if (multilines)
    return GTextRender::layoutText(f.value(), richtext, value(), -1, attrs, nullptr);
  else
    return f.value().textSize(value(), richtext, attrs);
}

// - - - -

ssize_t GText::posToIndex(ssize_t charpos, bool richtext) const {
  return strPosToIndex(value_, charpos, richtext);
}

ssize_t GText::indexToPos(ssize_t bytepos, bool richtext) const {
  return strIndexToPos(value_, bytepos, richtext);
}

ssize_t GText::nextIndex(ssize_t bytepos, bool richtext) const {
  return strNextIndex(value_, bytepos, richtext);
}

ssize_t GText::previousIndex(ssize_t bytepos, bool richtext) const {
  return strPreviousIndex(value_, bytepos, richtext);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GRegText::GRegText(GString const& value) : GText(value) {}

GRegText::GRegText(GRegText const& from) : GText(from.value_) {
  pattern_ = from.pattern_;
}

GRegText* GRegText::clone(bool copy_value) const {
  return copy_value ? new GRegText(*this) : new GRegText;
}

bool GRegText::isValid(GString const& s) const {
  if (pattern_.empty()) return true;
  std::regex pat{pattern_};
  return std::regex_match(s, pat);
}

GRegText& GRegText::operator=(GString const& v) {
  set(v,false);
  return *this;
}

GRegText& GRegText::operator=(const GText& v) {
  set(v.value(), false);
  return *this;
}

GRegText& GRegText::operator<<=(GExpr& expr) {
  GText::operator<<=(expr);
  return *this;
}

void GRegText::setFormat(GString const& pattern) {
  pattern_ = pattern;
}

void GRegText::setTimeFormat() {
  setFormat("^([0-9]|0[0-9]|1[0-9]|2[0-3]|):[0-5]?[0-9]?$");
}

void GRegText::setDateFormat() {
  setFormat("^(0?[1-9]|[12][0-9]|3[01])[\\-](0?[1-9]|1[012])[\\-]\\d{4}$");
}

void GRegText::setNameFormat() {
  setFormat("^[A-zÀ-ÖØ-öø-ÿ|\\-| |']*$");
}

void GRegText::setPhoneFormat() {
  setFormat("^[0-9|\\-|\\.| |']*$");
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
/*
GTitle* GTitle::clone(bool copy_value) const {
  return copy_value ? new GTitle(value_) : new GTitle;
}

GTitle& GTitle::operator=(GString const& v) {
  set(v,false);
  return *this;
}

GTitle& GTitle::operator=(const GText& v) {
  set(v.value(), false);
  return *this;
}

GTitle& GTitle::operator<<=(GExpr& expr) {
  GText::operator<<=(expr);
  return *this;
}

void GTitle::apply(Gadget* g, GRender& r, Specif specif) {
  // A REVOIR:
  // probleme: ces CSS n'excutent pas updateGadget(), il faut donc faire l'operation
  // dans apply() mais il faut ne le faire qu'une seule fois
  // ! ne marchera pas si le titre est partage par plusieurs gadgets vu
  // ! que isinit est propre à la prop
  //cerr << "GTitle::apply " <<" / "<< value_ << " / "<< int(objmodes_.isinit) <<endl;
  if (!omodes_.initialized) {
    omodes_.initialized = 1;
    if (g && g->toBox()) g->toBox()->setTitle(value_);
  }
}

 void GTitle::updateGadget(Gadget& g) {
 //if (g.toBox()) g.toBox()->setTitle(value_);
 }
 */

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GTip* GTip::clone(bool copy_value) const {
  return copy_value ? new GTip(*this) : new GTip;
}

bool GTip::onAdd(Gadget* g) {
  if (g) g->modes_.hasTip = 1;
  return true;
}

void GTip::onRemove(Gadget* g) {
  if (g) g->modes_.hasTip = 0;
}

void GTip::addModes(GadgetModes& m) {
  m.hasTip = 1;
}

void GTip::removeModes(GadgetModes& m) {
  m.hasTip = 0;
}

void GTip::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Tip]) return;
  r.specif[GPropType::Tip] = specif;
  r.tip_ = &value_;
}

}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
/*
 /// Title of a Gadget, a Box, a Window.
 class GTitle : public GText {
 public:
 GTitle(GString const& value = "") : GText(value) {}
 GTitle* clone(bool copy_value) const override;

 GTitle& operator=(GString const& val);
 GTitle& operator=(const GText&);

 using GText::operator<<=;
 GTitle& operator<<=(GExpr&);

 using GType = GPropType_<GTitle,GText>;
 static GType& Type();
 GType& type() const override {return Type();}

 void apply(Gadget*, GRender&, Specif) override;
 //void updateGadget(Gadget&) override;
 };

 /// Return a GTitle created in the heap.
 /// @{
 GTitle& Title();
 GTitle& Title(GString const& value);
 /// @}

 */
